from NightingaleORM.fields import Field
import datetime

class ModelMetaClass(type):
    def __new__(cls, name, bases, attrs):
        if name == "Model":
            return type.__new__(cls, name, bases, attrs)
        dateBaseType = attrs.get('__dbType__', '')  # 数据库类型
        dateBaseName = attrs.get('__dateBase__', '')  # 数据库名称
        schemaName = attrs.get('__schema__', '')  # 如果有schema，schema的名称
        tableName = attrs.get('__table__', None) or name.replace(
            'Model', '').lower()  # 表名称，如果没有默认实体去掉Model

        mappings = dict()  # 属性参照
        fields = []  # 字段
        primaryKey = None  # 主键

        for k, v in attrs.items():
            if isinstance(v, Field):
                mappings[k] = v
                mappings[k].__dateBase__=dateBaseName
                mappings[k].__schema__=schemaName
                mappings[k].__table__=tableName
                if v.primary_key:
                    if primaryKey:
                        raise RuntimeError(
                            "Douplicate primary key for field :%s" % primaryKey)
                    primaryKey = k
                else:
                    fields.append(k)
        if not primaryKey:
            raise RuntimeError("Primary key not found")
        # for k in mappings:
        #     attrs.pop(k)
        attrs['__mappings__'] = mappings  # 保存属性和列的映射关系
        attrs['__dateBase__'] = dateBaseName  # 保存db name
        attrs['__schema__'] = schemaName  # 保存schema
        attrs['__table__'] = tableName  # 保存table name
        attrs['__dbType__'] = dateBaseType  # 保存数据库类型
        attrs['__pk__'] = mappings[primaryKey]  # 保存数据库类型
        return type.__new__(cls, name, bases, attrs)

class BracketModel:
    '''
    括号整体类，括号的条件会关联在一起
    '''
    whereList=[]
    relation='AND'
    def __init__(self,model,relation):
        self.orginModel=model
        self.BracketModel=relation

    def addWhere(self,operation,relation='AND'):
        '''
        添加where条件
        operation:class ConditionModel or tuple like ('id','>',1)
        relation:'and' or 'or'
        '''
        if isinstance(operation,ConditionModel):
            operation.relation=relation
            self.whereList.append(operation)
        elif isinstance(operation,tuple) and len(operation)==3:
            con=ConditionModel(*operation)
            con.relation=relation
            self.whereList.append(con)
        return self
    
    def endBracket(self):
        self.orginModel.addBracketsWhere(self)
        return self.orginModel

class ConditionModel:
    fields='' #字段
    operation='=' #操作符
    value='' #值
    _relation='' #与前一个条件的关系

    def __init__(self,fields,operation,value):
        self.fields,self.operation,self.value=fields,operation.upper(),value 
    @property
    def relation(self): 
        return self._relation
    @relation.setter
    def relation(self,value):
        if isinstance(value,str):
            self._relation= value.upper()

class JoinConditionModel():
        joinType=''#the type of  join 
        onList=[] # the condition affter on 
        def __init__(self,joinType):
            self.joinType=joinType.upper()

        def addOn(self,condition,relation):
            condition.relation=relation
            self.onList.append(condition)

class OrderMOdel:
    field=''
    orderType='ASC'
    def __init__(self,field,orderType):
        self.field,self.orderType=field,orderType

class Model(dict, metaclass=ModelMetaClass):
    def __init__(self, **kw):
        super(Model, self).__init__(**kw)
        for k in self.__mappings__:
            field = self.__mappings__[k]
            if isinstance(field, Field):
                if field.name in self:
                    setattr(self, k, self[field.name])
                elif k in self:
                    setattr(self, k, self[k])
                else:
                    setattr(self, k, field.default)

    # 实现__getattr__与__setattr__方法，可以使引用属性像引用普通字段一样  如self['id']

    def __getattr__(self, key):
        try:
            if key not in self.__mappings__:
                raise AttributeError(r"'%s' is not field of db." % key)
            return self[key]
        except KeyError:
            raise AttributeError(r"'Model' object has no attribute '%s'" % key)

    def __setattr__(self, key, value):
        if key in  self.__mappings__:
            self[key]=Field.fieldTest(self.__mappings__[key],value)
        # self[key] = value

    __showFields__=[]
    __orderFields__=[]
    __whereFields__=[]
    __updateFields__=[]
    __bracketsWhereFields__=[]
    __Joins__=[]
    __alias__=''

    _interpreter=None#sql翻译器

    def addBracketsWhere(self,bracket:BracketModel):
        '''
        bracket:BracketModel
        '''
        if isinstance(bracket,BracketModel) and  len(bracket.whereList)>0:
            self.__bracketsWhereFields__.append(bracket)
        

    def addShow(self,*args):
        '''
        添加查询的fields
        '''
        self.__showFields__+=args
        return self
    
    def startBrackets(self,relation='AND'):
        '''
        开始括号 开始括号
        relation:
        '''
        return BracketModel(self,relation)
    
    def addWhere(self,operation,relation='AND'):
        '''
        添加where条件
        operation:class ConditionModel or tuple like ('id','>',1) or string
        relation:'and' or 'or'
        '''
        if isinstance(operation,ConditionModel):
            operation.relation=relation
            self.__whereFields__.append(operation)
        elif isinstance(operation,tuple) and len(operation)==3:
            con=ConditionModel(*operation)
            con.relation=relation
            self.__whereFields__.append(con)
        elif isinstance(operation,str):
             self.__whereFields__.append(operation)
        return self


    def addTableAlias(self,tableAlias):
        self.__alias__=tableAlias
        return self

    def addOrder(self,field):
        '''
        添加排序的的fields，升序
        '''
        self.__orderFields__.append(OrderMOdel(field,'ASC'))
        return self
    
    def addOrderDesc(self,field):
        '''
        添加排序字段，降序
        '''
        self.__orderFields__.append(OrderMOdel(field,'DESC'))
        return self


    def addJoin(self,*operation,joinType='JOIN'):
        '''
        添加Join条件
        model1 joinType model2 alias on operation
        '''
        join=JoinConditionModel(joinType)
        for item in operation:
            if isinstance(item,tuple):
                join.addOn(item[0],item[1])
            else:
                join.addOn(item,'AND')
        
        self.__Joins__.append(join)
        return self
    
    def addJoinStr(self,joinStr:str):
        '''
        添加Join条件
        joinStr
        '''
        if  isinstance(joinStr,str): 
            self.__Joins__.append(joinStr)
        return self
    
    def addUpdate(self,setCollection):
        '''
        添加Update条件
        set:the ConditionModel  .only equal  valid
        '''
        if isinstance(setCollection,ConditionModel):
            self.__updateFields__.append(setCollection)
        elif isinstance(setCollection,tuple) and len(setCollection)==3:
            con=ConditionModel(*setCollection)
            self.__updateFields__.append(con)
        elif isinstance(setCollection,str):
             self.__updateFields__.append(setCollection)
        return self

    def loadInterpreter(self,customModule):
        '''
        添加翻译其，默认自动识别，可手动传入，进行覆盖。不传递则自动识别
        add Interpreter ，default pgsql 
        '''
        self._interpreter=customModule

    def selectSql(self,count=0)->str:
        item=()
        myInterpreter=None
        if self._interpreter:
            myInterpreter=self._interpreter
        elif self.__dbType__.lower()=='pgsql':
            import NightingaleORM.pgsqlInterpreter as mpgsql
            myInterpreter=mpgsql
        item=myInterpreter.translateSelect(
            self.__dateBase__,
            self.__schema__,
            self.__table__,
            self.__alias__,
            self.__showFields__,
            self.__Joins__,
            self.__whereFields__,
            self.__bracketsWhereFields__,
            self.__orderFields__,
            self.__pk__,
            count,
            0)
        return item

    def pageOfList(self,index,pageSize):
        '''
        index:the index of pages
        pageSize:the size of each page
        '''
        myInterpreter=None
        if self._interpreter:
             # 这里是获取实际的值
            myInterpreter=self._interpreter
        elif self.__dbType__.lower()=='pgsql':
            import NightingaleORM.pgsqlInterpreter as mpgsql
            myInterpreter=mpgsql
        # 这里是获取实际的值
        item=myInterpreter.translateSelect(
                self.__dateBase__,
                self.__schema__,
                self.__table__,
                self.__alias__,
                self.__showFields__,
                self.__Joins__,
                self.__whereFields__,
                self.__bracketsWhereFields__,
                self.__orderFields__,
                self.__pk__,
                pageSize,
                (index-1)*pageSize)
        # 这里是获取总数的sql
        # item2=myInterpreter.translateSelect(
        #         self.__dateBase__,
        #         self.__schema__,
        #         self.__table__,
        #         self.__alias__,
        #         ['coungt(1)'],
        #         self.__Joins__,
        #         self.__whereFields__,
        #         self.__bracketsWhereFields__,
        #         self.__orderFields__,
        #         self.__pk__,
        #         pageSize,
        #         (index-1)*pageSize)    

        return item
    
    def insertSql(self):
        '''
        获取新增的的sql
        '''
        myInterpreter=None
        if self._interpreter:
            myInterpreter=self._interpreter
        elif self.__dbType__.lower()=='pgsql':
            import NightingaleORM.pgsqlInterpreter as mpgsql
            myInterpreter=mpgsql
        
        item=myInterpreter.translateInsert(
             self.__dateBase__,
            self.__schema__,
            self.__table__,
            self.__mappings__,
            self)
        return item

    def updateModel(self):
        '''
        获取新增的的sql
        '''
        myInterpreter=None
        if self._interpreter:
            myInterpreter=self._interpreter
        elif self.__dbType__.lower()=='pgsql':
            import NightingaleORM.pgsqlInterpreter as mpgsql
            myInterpreter=mpgsql
        item=myInterpreter.translateUpdateModel(
            self.__dateBase__,
            self.__schema__,
            self.__table__,
            self.__mappings__,
            self)
        return item

    def updateSql(self):
        item=()
        myInterpreter=None
        if self._interpreter:
            myInterpreter=self._interpreter
        elif self.__dbType__.lower()=='pgsql':
            import NightingaleORM.pgsqlInterpreter as mpgsql
            myInterpreter=mpgsql
        item=myInterpreter.translateUpdate(
            self.__dateBase__,
            self.__schema__,
            self.__table__,
            self.__updateFields__,
            self.__whereFields__,
            self.__bracketsWhereFields__)
        return item
    
    




from NightingaleORM import dbmodel
import re,datetime


class Field:
    def __init__(self, name, column_type, primary_key, default):
        self.name = name  # 字段名
        self.column_type = column_type  # 字段数据类型
        self.primary_key = primary_key  # 是否是主键
        self.default = default  # 有无默认值

    def __str__(self):
        return '<%s:%s>' % (self.__class__.__name__, self.name)

    def __lt__(self, rhs):
        return dbmodel.ConditionModel(self, '<', rhs)

    def __le__(self, rhs):
        return dbmodel.ConditionModel(self, '<=', rhs)

    def __gt__(self, rhs):
        return dbmodel.ConditionModel(self, '>', rhs)

    def __ge__(self, rhs):
        return dbmodel.ConditionModel(self, '>=', rhs)

    def __eq__(self, rhs):
        return dbmodel.ConditionModel(self, '=', rhs)

    def __ne__(self, rhs):
        return dbmodel.ConditionModel(self, '!=', rhs)

    @staticmethod
    def fieldTest(feild, value):
        '''
        return a ture value
        '''
        if value is None:
            return feild.default
        if isinstance(feild,StringField):
            if not isinstance(value, str):
                return feild.default
            containList = re.findall(r'[^()]+', feild.column_type)
            size = 0
            if len(containList) > 1:
                size = int(containList[1])
            if size > 0:
                if len(value) > size:
                    return value[:size]
                return value
            return value
        elif isinstance(feild,IntegerField):
            if not isinstance(value, int):
                return feild.default
            if 'int' in feild.column_type:
                if not (-2147483648 < value <2147483647 ):
                    return feild.default
            return value
        elif isinstance(feild,FloatField):
            if not isinstance(value, float):
                return feild.default
            return value
        elif isinstance(feild,DateTimeField):
            if isinstance(value, datetime.datetime):
                return value 
            if isinstance(value,str):
                try:
                    return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
                except :
                    pass
            return feild.default

        return value


class StringField(Field):
    def __init__(self, name=None, primary_key=False, default=None, ddl='varchar(128)'):
        super(StringField, self).__init__(name, ddl, primary_key, default)


class IntegerField(Field):
    def __init__(self, name=None, primary_key=False, default=None, ddl='bigint'):
        super(IntegerField, self).__init__(name, ddl, primary_key, default)


class FloatField(Field):
    def __init__(self, name=None, primary_key=False, default=None, ddl='float'):
        super(FloatField, self).__init__(name, ddl, primary_key, default)


class DateTimeField(Field):
    def __init__(self, name=None, primary_key=False, default=None, ddl='datetime)'):
        super(DateTimeField, self).__init__(name, ddl, primary_key, default)



from setuptools import setup, find_packages

setup(
    name='Nightingale-ORM',
    version='0.9.8',
    description=(
        '一个简单而小巧的orm工具，只用于生成sql和对象，而不依赖db驱动',
        'A simple and compact ORM tool that only generates SQL and objects, not DB drivers'
    ),
    long_description=open('README.rst','rb').read(),
    author='beincy',
    author_email='bianhui0524@sina.com',
    maintainer='卞辉(beincy)',
    maintainer_email='bianhui0524@sina.com',
    license='MIT',
    packages=['NightingaleORM'],
    platforms=["all"],
    url='https://github.com/beincy/Nightingale-ORM',
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)


from NightingaleORM.fields import Field
import json


def translateSelect(database, schema, table, alias, shows, joins, wheres, brackets, orders, pk, count=1, offset=0):
    if database is None or schema is None or table is None:
        raise 'interpreter cannot find the db or table'
    sqlStr = ''
    parameters = []
    if shows is None or len(shows) <= 0:
        raise 'select field can not be empty'
    # select
    sqlStrTop = f'''SELECT {translateShow(alias,shows)}'''
    sqlStrCountTop = 'SELECT count(1)'
    # from
    schemaStr = '' if len(schema) <= 0 else f'.{schema}'
    sqlStr = f'''{sqlStr} 
FROM {database}{schemaStr}."{table}" {alias}'''

    # join
    sqlStr = f'''{sqlStr} {translateJoin(joins,parameters)}'''
    # where
    if (wheres is None or len(wheres) <= 0) and (brackets is None or len(brackets) <= 0):
        raise 'where fields cannot be empty'

    sqlStr = f'''{sqlStr} 
WHERE 1=1{translateWhere(alias,wheres,parameters)}'''

    for bracketItem in brackets:
        sqlStr = f'''{sqlStr} {bracketItem.relation} (1=1 {translateWhere(alias,bracketItem.whereList,parameters)})'''
    # countsql
    sqlStrCountTop = f'''{sqlStrCountTop}{sqlStr}'''
    # order
    sqlStr = f'''{sqlStr} {translateOrder(orders,pk)}'''
    # limit
    if count>0:
        sqlStr = f'''{sqlStr} 
LIMIT {count}'''
    if offset>0:
        sqlStr = f'''{sqlStr}  OFFSET {offset}'''
    return (sqlStrTop+sqlStr, parameters, sqlStrCountTop)

# 添加where条件


def translateWhere(alias, wheres, parameters):
    sqlStr = ''
    for whereField in wheres:
        if isinstance(whereField, str):
            sqlStr = f'''{sqlStr} {whereField}'''
        else:
            sqlStr = f'''{sqlStr} {whereField.relation} {'' if len(alias)<=0 else f'{alias}.'}{whereField.fields if isinstance(whereField.fields,str) else whereField.fields.name}'''
            if whereField.operation == 'IN' or whereField.operation == 'NOT IN':
                if isinstance(whereField.value, list):
                    sqlStr = f'''{sqlStr} {whereField.operation} ('''
                    for whereItem in whereField.value:
                        parameters.append(whereItem)
                        sqlStr = f'''{sqlStr}${len(parameters)},'''
                    sqlStr = sqlStr.strip(',')
                    sqlStr = f'''{sqlStr})'''
                else:
                    sqlStr = f'''{sqlStr} {whereField.operation} ({whereField.value})'''
            elif whereField.operation == 'LEFT LIKE':
                parameters.append(f'%{whereField.value}')
                sqlStr = f'''{sqlStr} LIKE $%{len(parameters)}'''
            elif whereField.operation == 'RIGHT LIKE':
                parameters.append(f'{whereField.value}%')
                sqlStr = f'''{sqlStr} LIKE ${len(parameters)}'''
            elif whereField.operation == 'RIGHT LIKE':
                parameters.append(f'%{whereField.value}%')
                sqlStr = f'''{sqlStr} LIKE ${len(parameters)}'''
            else:
                parameters.append(whereField.value)
                sqlStr = f'''{sqlStr} {whereField.operation} ${len(parameters)}'''
    return sqlStr


def translateShow(alias, shows):
    sqlStr = ''
    for showfield in shows:
        sqlStr = f'''{sqlStr}{'' if len(alias)<=0 else f'{alias}.'}{ showfield.name if isinstance(showfield, Field) else showfield},'''
    sqlStr = sqlStr.strip(',')
    return sqlStr


def translateJoin(joins, parameters):
    sqlStr = ''
    if joins is not None and len(joins) > 0:
        sqlStr = ''
        for joinfield in joins:
            if isinstance(joinfield, str):
                sqlStr = f'''{joinfield[0]}'''
            else:
                onList = joinfield.onList
                if len(onList) > 0:
                    joinSchemaStr = ''
                    joinTable = ''
                    if all([onList[0], onList[0].fields]):
                        joinSchemaStr=f'.{onList[0].fields.__schema__}' if onList[0].fields.__schema__ else ''
                        joinTable = f'''{onList[0].fields.__dateBase__}{joinSchemaStr}."{onList[0].fields.__table__}" '''
                    sqlStr = f'''{sqlStr} 
{joinfield.joinType} {joinTable} ON'''
                    isFirst=True
                    for eachOn in onList:
                        if isinstance(eachOn, str):
                            sqlStr = f'''{sqlStr} {eachOn}'''
                        else:
                            if isinstance(eachOn.value, str):
                                parameters.append(eachOn.value)
                            else:
                                sqlStr = f'''{sqlStr} {'' if isFirst else eachOn.relation} {eachOn.fields.name} {eachOn.operation} {f'${len(parameters)}' if isinstance(eachOn.value,str) else eachOn.value.name}'''
                        isFirst=False
    return sqlStr


def translateOrder(orders, pk):
    sqlStr = ''
    if orders and len(orders) > 0:
        sqlStr = f'''
ORDER BY
        '''
        for order in orders:
            if isinstance(order, str):
                sqlStr = f'''{sqlStr} {order}'''
            else:
                sqlStr = f'''{sqlStr} {order.field.name} {order.orderType},'''
        sqlStr = sqlStr.strip(',')
    else:
        sqlStr = f'''
ORDER BY {pk.name}'''
    return sqlStr


def translateInsert(database, schema, table, mapDict, valueDict):
    '''
    get the sql of insert
    mapDict：Model_feild to  of db_feild
    valueDict:the model_feild Name
    '''
    if database is None or schema is None or table is None:
        raise 'interpreter cannot find the db or table'
    if mapDict is None or valueDict is None:
        raise AttributeError(r"db feild or value not be  none")

    schemaStr = '' if len(schema) <= 0 else f'.{schema}'
    insertStr = f'''INSERT INTO {database}{schemaStr}."{table}"'''  # 执行的sql
    parameters = []  # 参数化的参数
    fields = []  # 字段名称
    values = []  # 值
    for k, v in mapDict.items():
        if v.primary_key:
            continue
        if k in valueDict:
            parameters.append(valueDict.get(k))
            fields.append(v.name)
            values.append(f'${len(parameters)}')
        elif v.default is not None:
            parameters.append(v.default)
            fields.append(v.name)
            values.append(f'${len(parameters)}')
    if len(parameters) > 0:
        insertStr = f'''{insertStr} 
({','.join(fields)}) 
VALUES ({','.join(values)})'''
    return insertStr, parameters

def translateUpdateModel(database, schema, table, mapDict, valueDict):
    '''
    get the sql of UPDATE
    mapDict：Model_feild to  of db_feild
    valueDict:the model_feild Name
    '''
    if database is None or schema is None or table is None:
        raise 'interpreter cannot find the db or table'
    if mapDict is None or valueDict is None:
        raise AttributeError(r"db feild or value not be  none")

    schemaStr = '' if len(schema) <= 0 else f'.{schema}'
    updateStr = f'''UPDATE {database}{schemaStr}."{table}"'''  # 执行的sql
    parameters = []  # 参数化的参数
    fields = []  # 字段名称
    where = () # where条件
    for k, v in mapDict.items():
        if v.primary_key:
            if k not in valueDict  :
                raise 'primary_key not valid'
            if  isinstance(valueDict.get(k),int) and valueDict.get(k)<=0:
                raise 'primary_key not valid'
            where=(v.name,valueDict.get(k))
            continue
        if  k in valueDict:
            parameters.append(valueDict.get(k))
            fields.append(v.name)
        elif v.default is  not None:
            parameters.append(v.default)
            fields.append(v.name)
    if len(parameters) < 0:
         raise 'Update cannot find '
    if len(where)!=2:
         raise 'model not primary_key'
    updateStr = f'''{updateStr} 
SET'''
    for index,field in enumerate(fields):
        updateStr = f'''{updateStr} {field}=${index+1},'''
    
    updateStr = updateStr.strip(',')#去除最后一个空格

    #拼接 where条件
    parameters.append(where[1])
    updateStr = f'''{updateStr} 
WHERE {where[0]}=${len(parameters)}'''
    return updateStr,parameters

def translateUpdate(database, schema, table, updates,wheres, brackets):
    '''
    get the update sql 
    '''
    if database is None or schema is None or table is None:
        raise 'interpreter cannot find the db or table'
    if len(updates)<=0:
        raise 'updates must be any' 
    if len(wheres)<=0:
        raise 'wheres must be any' 
    schemaStr = '' if len(schema) <= 0 else f'.{schema}'
    sqlStr = f'''UPDATE {database}{schemaStr}."{table}"'''
    parameters = []
    sqlStr=f'''{sqlStr}
{translateSet(updates,parameters)}
WHERE 1=1{translateWhere('',wheres,parameters)}'''
    for bracketItem in brackets:
        sqlStr = f'''{sqlStr} {bracketItem.relation} (1=1 {translateWhere('',bracketItem.whereList,parameters)})'''
    return sqlStr,parameters

def translateSet(updates,parameters):
    sqlStr = ''
    if  updates and len(updates)>0:
        sqlStr = f'''SET '''
        for eachUpdate in updates:
            if isinstance(eachUpdate, str):
                sqlStr=f'''{sqlStr}{eachUpdate},'''
            else:  
                parameters.append(eachUpdate.value)
                sqlStr = f'''{sqlStr}{eachUpdate.fields.name} = ${len(parameters)},'''
    sqlStr = sqlStr.strip(',')
    return sqlStr
    